#include <set>
#include <iostream>
#include <algorithm>
#include "FrameSet.h"

/** Ctor with no windows at all.
*/
FrameSet::FrameSet() : wd(), areas()
{
}

/** Ctor with a single rectangular window.
* @param windo The sole window to represent the geometry.
*/
FrameSet::FrameSet(const Frame & windo)
{
	/* Insert the window into the list of all windows
	*/
	wd.push_back(windo) ;

	/* Starting index of flattened index
	*/
	areas.push_back(0) ;
	areas.push_back(windo.area()) ;

}


int FrameSet::covers( int x, int y) const
{
	for(int f=0 ;f < wd.size() ; f++)
		if ( wd[f].covers(x,y) )
			return f;
	return -1;
}

/**
* @param xedgvec
* @param yedgvec
* @param xidx
* @param yidx
* @return
*/
int FrameSet::covers(std::vector<int> &xedgvec, std::vector<int> & yedgvec, int xidx, int yidx) const
{
	/* The grid described by xedgvec and yedgvec describes 
	* rectangles with non-zero area, so checking the lower-left coordinates
	* for being covered is enough to decide where this area is covered.
	*/
	return covers(xedgvec[xidx],yedgvec[yidx]) ;
}

/** Compute the bounding box.
* @param bb On return, the values with the llx, lly, urx and ury integer
* coordinates of the maximum rectangular frame. The urx and ury coordinates
* are (in the usual C/C++ sense) the first integers outside any of the sub-frames.
*/
void FrameSet::boundbox(int bb[4]) const
{
	if ( wd.size() > 0)
	{
		/* initialize with the size of the first window.
		*/
		bb[0]= wd[0].posit.x ;
		bb[1]= wd[0].posit.y ;
		bb[2]= wd[0].posit.x +wd[0].nx ;
		bb[3]= wd[0].posit.y +wd[0].ny ;
		for(int f=1 ; f < wd.size() ; f++)
		{
			bb[0]= std::min(bb[0],wd[f].posit.x );
			bb[1]= std::min(bb[1],wd[f].posit.y );
			bb[2]= std::max(bb[2],wd[f].posit.x +wd[f].nx );
			bb[3]= std::max(bb[3],wd[f].posit.y +wd[f].ny );
		}
	}
	else
		bb[0]=bb[1]=bb[2]=bb[3] =0 ;
}

/** Merge the existing frames.
* The set of frames is split into non-overlapping frames, and these are
* merged into larger non-overlapping rectangular frames.
*/
void FrameSet::nonoverl()
{
	/* There is nothing to do if there is only one frame
	*/
	if ( wd.size() >1 )
	{

		/* The vertical and horizontal edges, 4 per frame, define a grid
		* on the x and y axes, which is used in a first run to cut
		* the existing frames into non-overlapping rectangular pieces.
		* Gather the information first in form of sets to eliminate stacked frames.
		*/
		std::set<int> xedg ;
		std::set<int> yedg ;
		for(int f=0 ; f < wd.size(); f++)
		{
			/* make all four corners of wd[f] known to the set of coordinates
			*/
			xedg.insert(wd[f].posit.x) ;
			xedg.insert(wd[f].posit.x + wd[f].nx) ;
			yedg.insert(wd[f].posit.y) ;
			yedg.insert(wd[f].posit.y + wd[f].ny) ;
		}
	
		/* The sorted variant: vectors. Reserve the space needed to
		* use the copy() of the algorithm library safely.
		*/
		std::vector<int> xedgvec(xedg.size()) ;
		std::vector<int> yedgvec(yedg.size()) ;
	
		std::copy( xedg.begin(), xedg.end(), xedgvec.begin() ) ;
		std::copy( yedg.begin(), yedg.end(), yedgvec.begin() ) ;
	

		std::sort( xedgvec.begin(), xedgvec.end() ) ;
		std::sort( yedgvec.begin(), yedgvec.end() ) ;


		/* This describes a irregular grid of verties
		*  (xedgvec[0],yedgvec[])  (xedgvec[1],yedgvec[])....  (xedgvec[],yedgvec[])
		*  ...
		*  (xedgvec[0],yedgvec[1])  (xedgvec[1],yedgvec[1])....  (xedgvec[],yedgvec[1])
		*  (xedgvec[0],yedgvec[0])  (xedgvec[1],yedgvec[0])....  (xedgvec[],yedgvec[0])
		* where the top row and right column describe the first pixel row/column outside the
		* topmost/rightmost frames. One can check for any individual sub-rectangle in that
		* grid whether this is either inside or outside of the frame set. In a simple
		* loop over these interior points, we can merge left-to-right (if both covered)
		* to reduce the number of rectangular frames.
		*/
		std::vector<Frame> mergefr ;
	
		/* loop over the rectangular grid. Since ydegvec[yedgved.size()-1] (and equivalent
		* the x coordinate) are the first pixels outside, the loops stop one index earlier
		* We maintain a flag to indicate whether the current status is being in an interstitial
		* (not covered) region and searching for the next covered frame, or wheter we are in
		* a merger mode trying to append more regions.
		*/ 
		Frame *newfr = 0 ;
		for(int yidx =0 ; yidx < yedgvec.size()-1 ; yidx++)
		{
			for(int xidx =0 ; xidx < xedgvec.size()-1 ; xidx++)
			{
				if ( covers(xedgvec,yedgvec,xidx,yidx) >= 0 )
				{
					const int dimx = xedgvec[xidx+1]-xedgvec[xidx] ;
					const int dimy = yedgvec[yidx+1]-yedgvec[yidx] ;
					if( newfr )
					{
						/* Glue the current region to the current frame.
						* The kind of walking ensures that this extends to the right (E).
						*/
						newfr->nx += dimx ;
					}
					else
					{
						/* Start a new frame
						*/
						newfr = new Frame(dimx, dimy, Point(xedgvec[xidx],yedgvec[yidx]) ) ;
					}
				}
				else
				{
					if(newfr)
					{
						/* Emit/save the current frame
						*/
						mergefr.push_back(*newfr) ;
						delete newfr ;
						newfr = 0 ;
					}
				}
			}
			if (newfr)
			{
				mergefr.push_back(*newfr) ;
				delete newfr ;
				newfr = 0 ;
			}
		}
		/* replace the current frame set with the merged one.
		*/
		wd = mergefr ;
	}

	/* update the hashed information on start index
	*/
	areas.clear() ;
	int flatidx = 0 ;
	for(int w=0 ; w < wd.size() ; w++)
	{
		areas.push_back(flatidx) ;
		flatidx += wd[w].area() ;
	}
	areas.push_back(flatidx) ;
}

/** Add the area of another rectangular frame to this set.
* @param windo The information on the additional rectangle covered.
*/
FrameSet  & FrameSet::operator+=(const Frame & windo)
{
	wd.push_back(windo) ;
	nonoverl() ;
	return *this ;
}

/** Add the areas of another set of frames to this .
* The rectangles defined by the argument are added to this set.
* The argument may contain partially overlapping areas. The internal
* representation is remodeled to describe a set of non-overlapping areas.
* @param wset The information on the additional rectangle covered.
*/
FrameSet  & FrameSet::operator+=(const FrameSet & wset)
{
	for(int w=0 ; w < wset.wd.size() ; w++)
		wd.push_back(wset.wd[w]) ;
	nonoverl() ;
	return *this ;
}

/** Total number of pixels squared.
* @return The sum over all areas of the subwindows.
* This is usually smaller than the prodcut of the edge lengths of the
* bounding box because here the interstitial area is not included.
*/
int FrameSet::area() const
{
	int ar = 0 ;
	/* Since the subwindows have been kept non-overlapping with all
	* operations, adding the partial areas up yields the total area.
	*/
	for(int w=0 ; w < wd.size() ; w++)
		ar += wd[w].area() ;
	return ar;
}

/** Match the index with a subframe number and 2D index in the subframe
* @param[in] idx The flattened index between 0 and the area.
* @param[out] widx The subframe widx[0], and the x and y coordinate within 
*  the subframe in widx[1..2].
*/
void FrameSet::loc(int idx, int widx[3]) const
{
	for(int f = 0 ; f < areas.size()-1; f++)
	{
		if ( areas[f+1] > idx)
		{
			widx[0] = f ;

			/* reduce idx to the value within window f
			*/
			idx -= areas[f] ;

			/* Integer division through edge length in x yields y index.
			* Remainder of that division is the x index.
			*/
			widx[2] = idx / wd[f].nx ;
			widx[1] = idx - wd[f].nx* widx[2] ;
			return ;
		}
	}
	/* TODO: error if out of range */
}


/** Write a describtion of a frame set to a stream.
* The format is (for each of the stripes) a literal f, a 0-based index,
* and the bracketed bounding box of the subwindow.
* @param os The stream the informatin is written to.
* @param fr The frame set to be described.
* @return
*/
std::ostream & operator<<(std::ostream & os, const FrameSet & fr)
{
	for (int f =0 ; f < fr.wd.size() ; f++)
		os << "f " << f << fr.wd[f] ;
	return os ;
}
